//
// INTEGR.CPP
//
// Modulo di integrazione e visualizzazione di CURVES 1.0
//
// Amato Gianluca, De Fanis Alberto, Paolini Emanuele
//

#include <alloc.h>
#include <math.h>
#include <graphics.h>
#include <float.h>

#include "vector.h"
#include "parser.h"
#include "integr.h"

//*** Variabili esterne

extern int maxx2,maxy2;           // getmaxx()/2 e getmaxy()/2
extern double mgx, mgy;           // numero di pixel per cm.

//*** Variabili locali

static int eye=0; 	 		      // 0:monocular 1:right -1:left

static vector3d x(1,0,0),y(0,1,0),z(0,0,1);
static vector3d def[38]={-x+z-y,x+z-y,x*3,x-y-z,x-y+z,-y*3,x-y-z,-x-y-z,
				-x-y+z,-y*3,-x-y-z,-z*3,x-y-z,x+y-z,x*3,x+y+z,x-y+z,z*3,
				-x-y+z,-x+y+z,x+y+z,z*3,-x+y+z,y*3,x+y+z,x+y-z,-z*3,-x+y-z,
				-x-y-z,-x*3,-x+y-z,x+y-z,y*3,-x+y-z,-x+y+z,-x*3,-x-y+z};

//*** Variabili pubbliche

double screenl=23;                // larghezza dello schermo in cm.
double kprosp=30;      			  // Distanza dello schermo dagli occhi cm.
double eyedist2=3.5; 	          // Distanza tra gli occhi/2 in cm.

unsigned int npunti=37;
vector3d *X=def;

Value mys("s",0,VARIABLE);
Expression *exprk, *exprt;

char WRONGPAR[]="parametri di integrazione fuori limiti";
static char SMALLDS[]="precisione troppo bassa";

//*** Chiamate parser

inline double k(double s)
{
	mys=s;
	return (*exprk)();
}

inline double tau(double s)
{
	mys=s;
	return (*exprt)();
}

//*** Integratore


// integra curve t e b con Runge-Kutta di grado 4

void integtb (vector3d &b, vector3d &t, double s, double ds)
{
	vector3d kt1,kt2,kt3,kt4,kb1,kb2,kb3,kb4,n,n1,n2,n3;

	n=b^t;
	kt1=n*k(s)*ds; kb1=n*(-tau(s))*ds; n1=(b+kb1/2)^(t+kt1/2);
	kt2=n1*k(s+ds/2)*ds; kb2=n1*(-tau(s+ds/2))*ds; n2=(b+kb2/2)^(t+kt2/2);
	kt3=n2*k(s+ds/2)*ds; kb3=n2*(-tau(s+ds/2))*ds; n3=(b+kb3)^(t+kt3);
	kt4=n3*k(s+ds)*ds; kb4=n3*(-tau(s+ds))*ds;

	t+=(kt1+kt2*2+kt3*2+kt4)/6;
	b+=(kb1+kb2*2+kb3*2+kb4)/6;
}


// integra s con corrector di grado 4

char *integrate (double s, double end, int num, int intmem)
{
	static allocated=0;

	vector3d vt[4], *ptx, xact(0,0,0), t(1,0,0), b(0,1,0);
	vector3d xeul;
	double ds;
	unsigned long i, maxmem;
	unsigned int old87=_control87(0,0), limit;
	char *er=0;

	if (num<=4 || intmem<=0)
		return WRONGPAR;
	maxmem=coreleft();
	if (maxmem>65520l) maxmem=65520l;
	if ((unsigned long)(num+1)*sizeof(vector3d)>maxmem)
		return NOMEM;
	ds=(end-s)/(num*intmem);
	if (ds==0) return WRONGPAR;      // ds troppo piccolo
	npunti=num;

	_control87 (0xFFFF,MCW_EM);
	if (allocated) delete X;
	X=new vector3d[npunti+1];
	allocated=1;
	noerror=1;

	// Inizializza integratore
	for (i=0, ptx=X; i<2; s+=ds, i++) {
		if ((i % intmem)==0)
			*(ptx++)=xact;
		vt[(int)i]=t;
		integtb (b,t,s,ds);
		xeul=xact+vt[(int)i]*ds;
		xact+=(vt[(int)i]+t)*ds/2;
		if (abs(xeul-xact)>abs(xact)/20)
			{er=SMALLDS;goto esci;}
		if ( _clear87() & (SW_OVERFLOW | SW_ZERODIVIDE | SW_INVALID) ) {
			er=WRONGEXPR; goto esci;
		}
	}

	limit=intmem*(npunti-1);
	for (vt[(int)i]=t; i<=limit; s+=ds, i++) {
		if ((i % intmem)==0)
			*(ptx++)=xact;
		integtb (b,t,s,ds);
		xeul=xact+(vt[(int)(i%4)]*23-vt[(int)((i-1)%4)]*16+vt[(int)((i-2)%4)]*5)*ds/12;
		xact+=(vt[(int)(i%4)]*19-vt[(int)((i-1)%4)]*5+vt[(int)((i-2)%4)]+t*9)*ds/24;
		vt[(int)((i+1)%4)]=t;
		if (abs(xeul-xact)>abs(xact)/20) {
			er=SMALLDS; goto esci;
		}
		if ( _clear87() & (SW_OVERFLOW | SW_ZERODIVIDE | SW_INVALID) ) {
			er=WRONGEXPR; goto esci;
		}

	}
  esci:
	if (er) {
		X=def; npunti=37;
	}
	noerror=0;
	_control87 (old87,MCW_EM);
	return er;
}

//*** Moduli visualizzazione

// conversione a coordinate video

inline int videox (double x) {
	return maxx2+(int)(((x)-eye*eyedist2)*mgx);
}

inline int videoy (double y) {
	return maxy2-(int)((y)*mgy);
}

// conversione 3D-2D

static char prospect (double &x, double &y, vector3d &X)
{
	if (X.x[2]<=1) return 0;  // Controllo visibilit
	x=kprosp*(X.x[0]+eye*eyedist2)/X.x[2];
	y=kprosp*X.x[1]/X.x[2];
	return 1;
}

void traslate(vector3d &t)
{
	for (vector3d *ptx=X; (ptx-X)<=npunti; ptx++)
		(*ptx)+=t;
}

void rotate(unsigned int d,double alpha)
{
	double x,y,c,s;
	unsigned int d1=(d+1)%3, d2=(d+2)%3;
	vector3d centre=*(X+npunti);

	c=cos(alpha); s=sin(alpha);
	traslate(-centre);
	for (vector3d *ptx=X; (ptx-X)<npunti; ptx++) {
		x=ptx->x[d1]*c-ptx->x[d2]*s;
		y=ptx->x[d1]*s+ptx->x[d2]*c;
		ptx->x[d1]=x;
		ptx->x[d2]=y;
	}
	traslate(centre);
}

void magnify(double k)
{
	vector3d centre=*(X+npunti);

	traslate(-centre);
	for (vector3d *ptx=X; (ptx-X)<npunti; ptx++)
		(*ptx)*=k;
	traslate(centre);
}

// calcola baricentro del punto

void addcentre()
{
	vector3d sum(0,0,0);
	for(vector3d *ptx=X; (ptx-X)<npunti; ptx++) {
		sum+=*ptx;
	}
	*ptx=sum/npunti;
}

// calcola diametro massimo curva

double diameter()
{
	double d=0,r;

	for (vector3d *ptx=X; (ptx-X)<npunti; ptx++)
		if ((r=abs(*ptx-X[npunti]))>d) d=r;
	return d*2;
}

// setta posizione e ingrandimento immagine di default

void initpos()
{
	vector3d t(0,0,kprosp);

	addcentre();
	traslate (t-X[npunti]);
	magnify (((double)maxy2)/(videoy(kprosp*diameter()/2/X[npunti].x[2])-maxy2+1));
}

//*** Visualizzazione

static void _draw()
{
	double x,y;

	for (vector3d *ptx=X; (ptx-X)<npunti; ptx++) {
		for (;(ptx-X)<npunti && !prospect(x,y,*ptx); ptx++);
			moveto(videox(x),videoy(y));
		for (;(ptx-X)<npunti && prospect(x,y,*ptx); ptx++) {
			asm cli
			lineto(videox(x),videoy(y));
			asm sti
		}
	}
}

void normal_draw()
{
	setcolor (WHITE);
	eye=0;
	_draw();
}

void stereo_draw()
{
	eye=1;
	setcolor(BLUE);
	_draw();
	eye=-1;
	setcolor(RED);
	setwritemode(OR_PUT);
	_draw();
	setwritemode(COPY_PUT);
}
