//
// KEYS.CPP
//
// Modulo gestione tasti di CURVES 1.0
//
// Amato Gianluca, De Fanis Alberto, Paolini Emanuele
//

#include <math.h>
#include <graphics.h>
#include <conio.h>
#include <string.h>
#include <ctype.h>

#include "keys.h"
#include "mouse.h"
#include "vector.h"
#include "integr.h"

//*** Variabili locali

static char stmouse=0;           // stato mouse


//*** Variabili esterne

extern int page;
extern char dualpage;
extern int maxx2,maxy2;
extern char self;
extern char redraw;
extern void menuview();
extern void draw();

//*** Routine di uso generale

// visualizza una stringa alla posizione corrente cancellando quella vecchia
// se del!=0 esegue solo la cancellazione.

void write(int x,int y,char *s,int del)
{
	int l=textwidth(s)+x-1+20, h=textheight(s)+y-1;
	int i=page;

	setcolor (0);
	setfillstyle (EMPTY_FILL,0);
	do {
		setactivepage(i);
		moveto(x,y);
		if (i==page) mouse_off();
		asm cli
		bar(x,y,l,h);
		if (!del) {
			setcolor (KEYCOLOR);
			outtext(s);
		asm sti
		}
		if (i==page) mouse_on();
	} while (dualpage && (i=1-i) != page);
}

// visualizza la stringa s alla posizione (x,y). cur  la posizione a cui
// visualizzare il cursore, -1 se no cursore.

void writecursor(int x,int y,char *s,int cur)
{
	int l,i=page;
	char c;

	write(x,y,s);
	if (cur==-1) return;
	c=*(s+cur);
	*(s+cur)=0;
	l=x+textwidth(s);
	*(s+cur)=c;
	do {
		setactivepage(i);
		moveto(l,y);
		if (i==page) mouse_off();
		asm cli
		lineto(l,y+textheight(s)-1);
		asm sti;
		if (i==page) mouse_on();
	} while (dualpage && (i=1-i) != page);
}

// inserisce un carattere all'inizio della stringa

void strins(char *s,char c)
{
	char t;
	for (; *s; s++) {
		t=*s; *s=c;	c=t;
	}
	*s=c;
	*(s+1)=NULL;
}

// cancella il primo carattere della stringa

void strdel(char *s)
{
	for (; *s; s++)
		*s=*(s+1);
}

// calcola la lunghezza di una stringa saltanto i ~

int mystrlen (char *s)
{
	int i=0;

	for (; *s; s++)
		if (*s!='~') i++;
	return i;
}


//*** Controllo background visualizzazione

// Restituisce tasti premuti

int mygetch()
{
	int x,y,c,count;
	static char precedence=0;
	static vector3d ztr(0,0,1);

	while (!kbhit()) {
		if (!stmouse) {
			mouse_getinfopressed (0,&count,&x,&y);
			stmouse=count!=0;
		} else {
			mouse_getinforeleased (0,&count,&x,&y);
			stmouse=count==0 ? 2 : 0;
		}
		if (!precedence && stmouse) {
			if (!count) { x=mouse_x(); y=mouse_y(); }
			y-=(maxy2*2-8*4);
			precedence=1;
			if (c=getkeyxy(NOKEY,x,y))
					return(c);
		}
		if (self) {
			rotate(1,ALPHA1);
			rotate(0,ALPHA2);
			traslate(ztr*TRASL);
			redraw=1;
		}
		draw(); precedence=0;
	}
	if (!(c=getch())) c=SPECIAL+getch();
	if (x=getkeyxy(c))
		return x;
	else
		return c;
}

// consente immissione di stringhe

char *getline(char *s,int n)
{
	int x,y;  			//coordinate punto iniziale di scrittura
	int cur=0;    		//posizione cursore
	int ch;
	static int GETKEYS[]={ESC,CTRLBACKSPACE,0};

	x=getx(); y=gety();
	cur=strlen(s);
	keyoff(NULL,0);
	keyon(GETKEYS);
	while (1) {
		menuview();
		writecursor(x,y,s,cur);
		switch(ch=mygetch()) {
			case LEFT:
				if (cur>0) cur--;
				break;
			case RIGHT:
				if (cur<strlen(s)+1) cur++;
				break;
			case DEL:
				if (cur<strlen(s)) strdel(s+cur);
				break;
			case BACKSPACE:
				if (cur>0) strdel(s+(--cur));
				break;
			case ESC:
			case ENTER:
				keyoff(GETKEYS);
				return(s);
			case CTRLBACKSPACE:
				strset (s,' '); menuview();
				writecursor (x,y,s,-1);
				*s=0; cur=0;
				break;
			case NOKEY:
			default:
				if (isprint(ch) && strlen(s)<n-1) strins(s+(cur++),ch);
				break;
		}
	}
}


//*** Classe tasto

Key* Key::first=NULL;

Key::Key(char *s,int k,char type, int kx,int ky,int w,int h)
{
	x=kx,y=ky; key=k; name=s;
	if (w==0)
		w=mystrlen(s)*8+10;
	if (h==0)
		h=textheight(s)+4;
	if (type & KXCENTER) x-=w/2;
	if (type & KYCENTER) y-=h/2;
	xx=x+w; yy=y+h;
	ison=0;
	repeat=(type & KREPEAT)!=0;
	if (type & KDELETE) off();
	next=first;
	first=this;
}

Key::~Key()
{
	Key *p;
	if (ison) off();
	if (first==this) {
		first=this->next;
	} else {
		for (p=first; p->next!=this; p=p->next);
		p->next=this->next;
	}
}

void Key::on()
{
	int i=page,xt;
	char *p1,*p2;

	if (p1=strchr(name,'~')) *p1=0;
	if (p2=strchr(p1+1,'~')) *p2=0;
	menuview();
	setcolor(KEYCOLOR);
	do {
		setactivepage (i);
		if (i==page) mouse_off();
		asm cli
		rectangle(x,y,xx,yy);
		moveto (x+5,y+(yy-y-textheight(name))/2);
		outtext (name);
		if (p1) {
			xt=getx();
			outtext(p1+1);
			line (xt,yy-2,getx(),yy-2);
			if (p2)	outtext (p2+1);
		}
		asm sti
		if (i==page) mouse_on();
	} while (dualpage && (i=1-i) != page);
	if (p1) *p1='~';
	if (p2) *p2='~';
	ison=1;
}

void Key::off()
{
	int i=page;

	menuview();
	setfillstyle(EMPTY_FILL,0);
	do {
		setactivepage(i);
		if (i==page) mouse_off();
		asm cli
		bar(x,y,xx,yy);
		asm sti
		if (i==page) mouse_on();
	} while (dualpage && (i=1-i) != page);
	ison=0;
}

int Key::touch(int kx,int ky, int c)
{
	return( ison && (c==(key & MASKBIT) || (stmouse==1 || repeat) && kx>=x
			&& kx<=xx && ky>=y && ky<=yy));
}

//*** Classe tasto di immissione stringhe

InputKey::InputKey(char *s,int k,int l,int kx,int ky,int w,int h):
					Key(s,k,0,kx,ky,w,h)
{
	val=new char[len=l];

}

void InputKey::off()
{
	xx+=8*len;
	Key::off();
	xx-=8*len;
}

void InputKey::on()
{
	menuview();
	setcolor(KEYCOLOR);
	if (!ison) off();
	Key::on();
	write(xx+5,y+2,val);
}

InputKey::touch(int kx, int ky, int c)
{
	if (ison && (stmouse==1 && kx>=x && kx<=xx+8*len && ky>=y && ky<=yy
		 || (kx==-1) || (key & MASKBIT)==c)){
		moveto(xx+5,y+2);
		getline(val,len);
		int t=ison;
		off();
		on();
		ison=t;
		return 1;
	}
	return 0;
}

//*** Classe trimmer a scorrimento

Trimmer::Trimmer (char *s,int k,int kx,int ky,int w,int h):
				  Key(s,k,KREPEAT,kx,ky,w,h) {val=0;}

void Trimmer::on()
{
	int i=page;

	menuview();
	setcolor(KEYCOLOR);
	do {
		setactivepage(i);
		if (i==page) mouse_off();
		asm cli
		rectangle(x,y,xx,yy);
		outtextxy((x+xx)/2-textwidth(name)/2,yy-(h=textheight(name)),name);
		rectangle(x+2,y+2,x+12,yy-h-2);
		rectangle(xx-12,y+2,xx-2,yy-h-2);
		rectangle(x+14,y+2,xx-14,yy-h-2);
		asm sti
		if (i==page) mouse_on();
	} while (dualpage && (i=1-i) != page);
	ison=1; touch();
}

int Trimmer::touch(int kx, int ky, int c)
{
	int i=page;

	if (ison && (key & MASKBIT)==c)
		return 1;
	if (ison && (kx>x && ky>y && kx<xx && ky<yy) || kx==-1) {
		if ( kx!=-1 ) {
			if (kx<x+12)
				val-=1.0/(xx-x-30);
			else if (kx>xx-12)
				val+=1.0/(xx-x-30);
			else if (ky>yy-10)
				val=0;
			else
				val=((double)(kx-(xx+x)/2)/((xx-x-28)/2));
			}
		if (val<-1) val=-1;
		if (val>1) val=1;

		menuview();
		setfillstyle (EMPTY_FILL,0);
		setcolor (KEYCOLOR);
		do {
			setactivepage(i);
			if (i==page) mouse_off();
			asm cli
			bar(x+15,y+3,xx-15,yy-h-3);
			line((xx+x)/2+(xx-x-30)*val/2,y+2,(xx+x)/2+(xx-x-30)*val/2,yy-h-2);
			asm sti
			if (i==page) mouse_on();
		} while (dualpage && (i=1-i) != page);
		return(1);
	}
	return(0);
}

//*** Funzioni gestione generale tasti

int getkeyxy(int c,int x,int y)
{
	Key *p;

	for (p=Key::first; p; p=p->next)
		if (p->touch(x,y,c)) return (p->key & MASKBIT);

	return stmouse==1 ? NOKEY : 0;
}

Key *getkey(int c)
{
	Key *p;
	for (p=Key::first;p;p=p->next)
		if (p->key==c) return p;
	return(NULL);
}

int strintr(int *p, int d)
{
	while (*p)
		if (d==*(p++)) return 1;
	return 0;
}

void keyon(int *s, int draw) // s=NULL -> rewrite keys
{
	Key *p;

	for (p=Key::first; p; p=p->next)
		if ((s==NULL && p->ison) || (s && strintr(s,p->key)))
			if (draw) p->on(); else p->ison=1;
}

void keyoff(int *s, int draw)  //s=NULL -> ALL
{
	Key *p;

	for (p=Key::first;p; p=p->next)
		if ((s==NULL || strintr(s,p->key))&&p->ison)
			if (draw) p->off(); else p->ison=0;
}

