//
// PARSER.CPP
//
// Modulo per interpretazione e compilazione di espressioni matematiche
// versione 1.0
//
// Amato Gianluca, De Fanis Alberto, Paolini Emanuele
//

#include <ctype.h>
#include <iostream.h>
#include <stdlib.h>
#include <math.h>
#include <dos.h>
#include <float.h>

#include "parser.h"

//*** Gestione errori

#define _ITALMSG_

static char * error=NULL;   // contiene il tipo di errore, NULL nessun errore
static int poserror=0;      // posizione errore nella stringa di input

int noerror=0;       		// se 1 matherr non da mai errore.

#ifdef _ITALMSG_

char NOMEM[]="memoria insufficiente";
char WRONGEXPR[]="l'espressione ha condotto ad una operazione illegale";

static char RBRACKEXPECTED[]="si aspettava una ')'";
static char ENDEXPECTED[]="si aspettava la fine dell'espressione";
static char UNKNOWN[]="simbolo sconosciuto";
static char OVERSTACK[]="formula troppo complessa";
static char NOTEXPECTED[]="qualcos'altro  richiesto in questo punto";
static char DUMMY[]="";

#else

char NOMEM[]="insufficient memory";
char WRONGEXPR[]="expression lead to illegal operations";

static char RBRACKEXPECTED[]="')' expected";
static char ENDEXPECTED[]="end of expression expected";
static char UNKNOWN[]="unknown symbol";
static char OVERSTACK[]="formula too complex";
static char NOTEXPECTED[]="something other is expected here";
static char DUMMY[]="";

#endif

// ridefinizioe della matherr standard per evitare che il programma si
// interrompa durante la compilazione

int matherr(struct exception *e)
{
	if (noerror) {
		error=WRONGEXPR;
		return 1;
	} else
		if ( (e->type==UNDERFLOW) || (e->type==OVERFLOW) )
			return 1;
		else
			return 0;
}

// restituisce l'errore corrente

int parsererror(char **reterror)
{
	int temp=poserror ? poserror : -(error!=NULL);

	*reterror=error;
	error=NULL; poserror=0;
	return temp;
}

//*** Funzioni per la compilazione

func funcptr;

static int busy; 	         // quanti valori ci sono nello stack;
static int maxbusy;          // valore massimo occupazione stack
static unsigned int currseg; // segmento corrente
static int statecomp;        // stato di compilazione corrente
static int lencode;          // lunghezza codice attuale

static void pushint (unsigned int i)
{
	if (statecomp) *funcptr++=i;
	lencode+=2;
}

static void pushchar (char i)
{
	if (statecomp) *(((char *)funcptr)++)=i;
	lencode+=1;
}

static void pushistr (unsigned int i)
{
	if (_8087)
		pushint (i);
	else {
		pushchar (0xCD);
		pushchar ((char)i-0xD8+0x34);
		pushchar (*((char*)(&i)+1));
	}
}

static void pushdata(double far *p)
{
	busy++;
	if (busy>maxbusy) maxbusy=busy;
	if (currseg!=FP_SEG(p)) {
		pushchar (0xB8);
		pushint (currseg=FP_SEG(p));  // moc ax,seg(*p)
		pushint (0xD88E);             // mov ds,ax
	}
	pushistr(0x06DD);
	pushint(FP_OFF(p));               // fld (*p)
}

static void pushconst(double far *p)
{
	pushdata(p);
}

static void pushvar(double far *p)
{
	pushdata (p);
}

static void c_sum(int)
{
	busy--;pushistr (0xC1DE); // faddp
}

static void c_sub(int rev)
{
	busy--; pushistr (rev ? 0xE1DE : 0xE9DE); // fsubrp o fsubp
}

static void c_mul(int)
{
	busy--; pushistr (0xC9DE); // fmulp
}

static void c_div(int rev)
{
	busy--;	pushistr (rev ? 0xF1DE : 0xF9DE); // fdivrp o fdivp
}

static void c_2x()
{
	static int expcode[]={0xC0D9, 0xFCD9, 0xC9D9, 0xE1D8, 0xF0D9,
						  0xE8D9, 0xC1DE, 0xFDD9, 0xC1DD};
	// fld st(0); frndint; fxch st(1); fsub st,st(1); f2xm1
	// fld1; faddp; fscale; ffree st(1)
	if (busy+1>maxbusy) maxbusy=busy+1;
	for (int i=0; i<=8; i++) pushistr (expcode[i]);
}

static void c_pow(int rev)
{
	static int powcode[]={0xE8D9, 0xC9D9, 0xF1D9, 0xC9DE};
	// fld1; fxch st(1); fyl2x; fmulp;
	if (busy+1>maxbusy) maxbusy=busy+1; busy--;
	if (!rev) pushistr (0xC9D9); // fxch st(1)
	for (int i=0; i<=3; i++) pushistr(powcode[i]);
	c_2x();
}

static void c_chs(int)
{
	pushistr (0xE0D9); // fchs
}

static void c_exp(int)
{
	if (busy+1>maxbusy) maxbusy=busy+1;
	pushistr (0xEAD9); pushistr (0xC9DE); // fldl2e; fmulp
	c_2x();
}

static void c_ln(int)
{
	static int lncode[]={0xE8D9, 0xC9D9, 0xF1D9, 0xEAD9, 0xF9DE};
	// fld1; fxch st,st(1); fyl2x; fldl2e; fdivp
	if (busy+1>maxbusy) maxbusy=busy+1;
	for (int i=0; i<=4; i++) pushistr (lncode[i]);
}

static void c_sqr(int)
{
	pushistr (0xC8D8); // fmul st,st(0)
}

static void c_sqrt(int)
{
	pushistr (0xFAD9); // fsqrt
}

static void c_sin(int)
{
	if (_8087>=3)
		pushint (0xFED9); // fsin
	else {
	  if (busy+2>maxbusy) maxbusy=busy+2;
	  pushint (0x3ECD); pushint (0x90EC); // call fsin
	}
}

static void c_cos(int)
{
	if (_8087>=3)
		pushint (0xFFD9); // fcos
	else {
		if (busy+2>maxbusy) maxbusy=busy+2;
		pushint (0x3ECD); pushint (0x90EE); // call fcos
	}
}

static void c_tan(int)
{
	if (busy+1>maxbusy) maxbusy=busy+1;
	if (_8087>=3) {
	  pushint (0xFBD9); pushint (0xF9DE); // fsincos; fdivp
	} else {
	  pushint (0x3ECD); pushint (0x90F0); // call ftan
	}
}

static void c_atan(int)
{
	if (busy+1>maxbusy) maxbusy=busy+1;
	pushint (0x3ECD); pushint (0x90F2); // call atan
}

//*** Funzioni per la classe base e Object

static void mydelete(Base *x)
{
	if ( x->deletable ) delete(x);
}

ostream &operator << (ostream &out,Base &o)
{
	return(o.out(out));
}

int lastprior=0;      //priorit massima attuale

//*** Classe costanti numeriche

class Number:public Object
{
		double n;
		virtual ostream &out(ostream &out) { return out<<n; };
	public:
		Number(double x) { deletable=1; n=x; };
		virtual int evalue(double *r) { *r=n; return 1;};
		virtual void compile() { pushconst(&n); };
};

//*** Classi funzioni, operatori, costanti

typedef double(*myfunc)(double,double);

class Function:public Identifier  //sqrt, sin ,cos, +, -, *...
{
	public:
		double (*fun)(double,double);
		void (*c_fun)(int);
		Function(char *s,myfunc f,void(*c_f)(int),unsigned int n):
			Identifier (s, n|FUNCTION) { fun=f; c_fun=c_f;};
};

class Operator:public Function
{
		int pri;

	public:
		virtual int priority(){return pri;};
		Operator (char *s,myfunc f, void(*c_f)(int),unsigned int n,int ass=0);
};

Operator::Operator (char *s,myfunc f,void(*c_f)(int),unsigned int n,int ass):
					Function (s, f, c_f, n|OPERATOR)
{
	pri=ass;
	if (ass>lastprior) lastprior=ass;
}

int Value::evalue(double *r)
{
	if (type & CONSTANT) {
		*r=v; return(1);
	}
	return(0);
}

void Value::compile()
{
	if (type & CONSTANT)
		pushconst(&v);
	else
		pushvar(&v);
}

//*** Funzioni per la valutazione

static double sum(double x,double y) { return(x+y); }
static double sub(double x,double y) { return(x-y); }
static double mul(double x,double y) { return(x*y); }
static double div(double x,double y) { return(x/y); }
static double chs(double x,double ) { return(-x); }
static double mysqr(double x) { return(x*x); }

//*** Identificatori di default

static Operator
	SUM("+",sum,c_sum,BINARY,1),
	SUB("-",sub,c_sub,BINARY,1),
	MUL("*",mul,c_mul,BINARY,2),
	MUL2("",mul,c_mul,BINARY,2),
	DIV("/",div,c_div,BINARY,2),
	POW("^",pow,c_pow,BINARY | RIGHTASSOCIATION,3),
	CHS("-",chs,c_chs,UNARY,4);

static Function
	_1("sqrt",(myfunc)sqrt,c_sqrt,UNARY),
	_2("sin",(myfunc)sin,c_sin,UNARY),
	_3("cos",(myfunc)cos,c_cos,UNARY),
	_4("exp",(myfunc)exp,c_exp,UNARY),
	_5("ln",(myfunc)log,c_ln,UNARY),
	_6("tan",(myfunc)tan,c_tan,UNARY),
	_7("sqr",(myfunc)mysqr,c_sqr,UNARY),
	_8("atan",(myfunc)atan,c_atan,UNARY);

static Value
	_PI("PI",M_PI,CONSTANT);

//*** Opzioni di visualizzazione

VIDMODE vidmode=INFISSA;

//*** Classe Tree

class Tree:public Object
{
		static Object Foglia;
		Function *it;
		Object *left;
		Object *right;
		char reverse;
		virtual ostream& out(ostream &);

	public:
		Tree(Function *f=NULL):Object()
			{ deletable=1; reverse=0; it=f; left=&Foglia; right=&Foglia; };
		virtual ~Tree() { mydelete (it); mydelete(left); mydelete(right); };
		virtual int priority() { return it->priority(); };
		virtual int simplify_length();
		virtual Object *simplify_const();
		virtual int evalue(double *r);
		virtual void compile();
		friend class Expression;
};

Object Tree::Foglia;

void Tree::compile()
{
	if (reverse) {
		right->compile(); left->compile();
	} else {
		left->compile(); right->compile();
	}
	it->c_fun(reverse);
}

Object *Tree::simplify_const()
{
	double val;
	Object *p;

	left=left->simplify_const();
	right=right->simplify_const();
	if ( evalue(&val) )
	{
		p=(Object*)(new Number(val));
		if (p) {
			mydelete (this); return p;
		}
	}
	return this;
}

int Tree::simplify_length()
{
	int a,b;

	a=left->simplify_length();
	if (it->type & UNARY)
		return a;
	else {
		b=right->simplify_length();
		if (b>a) reverse=!reverse;
		return( (a>b ? a : b)+1 );
	}
}

int Tree::evalue(double *r)
{
	double leftval,rightval;

	if ( left->evalue(&leftval) )
	{
		if (it->type & BINARY )
			if ( right->evalue(&rightval) )
				*r=it->fun(leftval,rightval);
			else
				return(0);
		else
			*r=it->fun(leftval,0.0);
		return 1;
	}
	return(0);
}

static int brackets (Function *it, Object *arg, int is_left)
{
	if (it->priority() == arg->priority())
		return (! (is_left ^ !! (it->type & RIGHTASSOCIATION)));
	return( arg->priority() < it->priority() );
}

ostream& Tree::out(ostream & out)
{
	if (vidmode==RPN)
		if (it->type & BINARY)
			return(out << *left << ' ' << *right << ' ' << *it);
		else
			return(out << *left << ' ' << *it);
	else {
		if (it->type & BINARY) { //si suppone sia un operatore
			if ( brackets (it,left,1))
				out <<'(' << *left <<')';
			else
				out << *left;
			out << *it;
			if (brackets(it,right,0))
				out<< '(' << *right <<')';
			else
				out << *right;
		} else if (it->type & OPERATOR )  //ad esempio meno unario
			out << *it << *left;
		else //funzione
			out << *it << '(' << *left<< ')';
	}
	return(out);
}

//*** Classe Expression

int Expression::installide (Identifier *p)
{
	node *n=new node;

	if (!n) {
		error=NOMEM; return 0;
	}
	n->link=p; n->next=firstidentifier;
	firstidentifier=n;
	return 1;
}

int Expression::installdefault ()
{
	Identifier *arr[]={&SUM, &SUB, &MUL, &MUL2, &DIV, &POW, &CHS, &_1,
		&_2, &_3, &_4, &_5, &_6, &_7, &_8, &_PI};
	int res=1;

	for (int i=0; i<=15 && res; i++) res=installide(arr[i]);
	return res;
}

void Expression::onlycompile()
{
	// stima lunghezza codice
	busy=maxbusy=0; lencode=0;
	statecomp=0; currseg=_DS;
	ptr->compile();
	funcpoint=funcptr=(func) new char[lencode+1];
	if (!funcpoint) {
		error=NOMEM; return;
	}
	if (maxbusy>8) {
		error=OVERSTACK; return;
	}
	// compila effettivamente
	#ifdef __PARDEBUG__
	cout << "END. \n";
	#endif
	currseg=_DS; statecomp=1;
	ptr->compile();
	#ifdef __PARDEBUG__
	cout << "END. \n";
	#endif
	pushchar (0xCB);
}

#pragma warn -rvl
double Expression::operator ()()
{
	func jumpit=funcpoint;
	asm {
		push ds
		call dword ptr jumpit
		pop ds
		fwait
	};
	#if __BORLANDC__==0x200
	if (0) return(0);
	#endif
}
#pragma warn +rvl

static char *blank(char *&s)
{
	while (isspace(*s)) s++;
	return(s);
}

static int strcmplen(char *s,char *t) //s comincia per t?
{
	for(int i=0; *t && (*s==*t); t++,s++,i++);
	if (*t==0)
		return(i);
	else
		return(-1);
}

Identifier *Expression::ident_expr(int prior, int type)
{
	int maxlen=-1,n;
	node *a;
	Identifier *p,*id;

	blank(s);
	if (! (*s) || (*s==')') ) {
		error=NOTEXPECTED; return NULL;
	}
	for (a=firstidentifier; a!=NULL; a=a->next)
	{
		p=a->link;
		if ( (p->type & type)==type && (n=strcmplen(s,p->name)) > maxlen )
		{
			maxlen=n; id=p;
		}
	}
	if (maxlen>=0) {
		if (id->priority()==prior) {
			s+=maxlen;
			blank(s);
		} else {
			error=DUMMY; id=NULL;   // identificatore di priorit diff.
		}
	} else {
		error=UNKNOWN; id=NULL;     // nessun identificatore
	}
	return id;
}

Object *Expression::base_expr()
{
	Object *expr;

	blank (s);
	if (*s=='(') {
		s++;
		if ( !(expr=MakeTree (1)) ) return NULL;
		blank (s);
		if (*s!=')') {
			mydelete(expr); error=RBRACKEXPECTED; return NULL;
		}
		s++;
		return expr;
	} else if (isdigit(*s) || *s=='.' && isdigit(*(s+1)) ) {   //  un numero
		if (! (expr=(Object*)new Number(strtod(s,&s))) ) {
			error=NOMEM; return NULL;
		}
		return expr;
	} else {			// un identificatore
		#pragma warn -pia
		if ( expr=(Identifier *)ident_expr(lastprior,FUNCTION) ) {
		#pragma warn +pia
			Object *arg;
			Tree *tree;

			if ( ! (arg=base_expr()) ) {
				mydelete (expr); return NULL;
			}
			if ( ! (tree=new Tree) ) {
				error=NOMEM; mydelete (expr); mydelete (arg);
				return NULL;
			}
			tree->it=(Function*)expr;
			tree->left=arg;
			return(tree);
		}
		if ( !(expr=(Identifier *)ident_expr(lastprior,VALUE)) )
			return NULL;
		else
			return expr;
	}
}

Object* Expression::MakeTree (int prior)
{
	Identifier *op;
	Tree *tree;
	Object *left,*right;

	if ( prior>=lastprior )
		return base_expr();
	op=ident_expr(prior,UNARY);
	if (op) {
		if ( ! (left=MakeTree(prior)) ) return NULL;
		if (! (tree=new Tree((Function *)op)) ) {
			mydelete (left); error=NOMEM; return NULL;
		}
		tree->left=left;
		return(tree);
	}
	error=NULL;
	if ( ! (left=MakeTree(prior+1)) ) return NULL;
	while (1) {
		if (! (op=ident_expr(prior,OPERATOR | BINARY))) {
			error=NULL; return(left);
		}
		if (op->type & RIGHTASSOCIATION)
		{
			if ( ! (right=MakeTree(prior)) ) {
				mydelete(left); return NULL;
			}
			if (! (tree=new Tree((Function *)op)) ) {
				mydelete (left); mydelete(right);
				error=NOMEM; return NULL;
			}
			tree->left=left;
			tree->right=right;
			return(tree);
		}
		if (! (right=MakeTree(prior+1)) ) {
			mydelete (left); return NULL;
		}
		if ( ! (tree=new Tree((Function *)op)) ) {
			mydelete(left); mydelete(right);
			error=NOMEM; return NULL;
		}
		tree->left=left;
		tree->right=right;
		left=tree;
	}
}

Expression::Expression(char *str, Value *var)
{
	firstidentifier=NULL; funcpoint=NULL;
	ptr=&Tree::Foglia; s=sold=str;
	if (installdefault() && var)
		installide (var);
}

void Expression::maketree()
{
	ptr=MakeTree(1);
	if (ptr && *s) {
		error=ENDEXPECTED; delete ptr; ptr=&Tree::Foglia;
	} else if (!ptr)
		ptr=&Tree::Foglia;
	if (error) poserror=(int)(s-sold+1);
}

void Expression::simplify_const()
{
	unsigned int old87=_control87(0,0);

	noerror=1;
	_control87 (0xFFFF,MCW_EM);
	ptr=ptr->simplify_const();
	noerror=0;
	if ( _clear87() & (SW_ZERODIVIDE | SW_OVERFLOW | SW_INVALID) )
		error=WRONGEXPR;
	_control87 (old87,MCW_EM);
}

void Expression::compile()
{
	maketree();
	if (error) return;
	simplify_const();
	if (error) return;
	simplify_length();
	onlycompile();
}
