//
// PARSER.H
//
// Modulo per interpretazione e compilazione di espressioni matematiche
// versione 1.0
//
// Amato Gianluca, De Fanis Alberto, Paolini Emanuele
//

#ifndef __PARSER__

#define __PARSER__

//*** Gestione errori

int parsererror (char **error);

extern char NOMEM[];
extern char WRONGEXPR[];
extern int noerror;

typedef unsigned int far * func;   // tipo della funzione compilata

//*** Classe base e object

class Base
{
		virtual ostream &out(ostream &out) { return(out<<"ITEM"); };
		friend void mydelete(Base *x);

	public:
		int deletable;
		Base() { deletable=0; };
		virtual ~Base() {};
		friend ostream &operator << (ostream &,Base &); //output dell'oggetto
};

extern int lastprior;      //priorit massima attuale

class Object:public Base   //classe di base per tutti gli oggetti
{
	public:
		// semplifica lunghezza rami
		virtual int simplify_length() { return 0; };

		// semplifica operazioni su costanti
		virtual Object* simplify_const() { return this; };

		virtual int priority() { return lastprior; };
		virtual int evalue(double *) { return 0;} ;
		virtual void compile() {};  // compila l'oggetto
};

//*** Identificatori, valori

// tipo identificatori

#define VOID     0
#define FUNCTION 1   // una funzione
#define OPERATOR 2   // un operatore
#define UNARY    4   //funzione/operazione unaria(prefissa)
#define BINARY   8   //funzione binaria (o operatore binario)
#define VALUE    16
#define VARIABLE 32
#define CONSTANT 64
#define RIGHTASSOCIATION 128 //ON se associa a destra

class Identifier:public Object
{
		ostream &out(ostream &out) { return out<<name; };

	public:
		char* name;  		 //stringa che contiene l'identificatore
		unsigned int type;   //vedi define precedente
		virtual int evalue(double *) { return 0; };
		Identifier (char *s=NULL,unsigned int t=VOID) { name=s; type=t; };
};

class Value:public Identifier //x,y,PI ...
{
		double v;

	public:
		Value(char*s,double val=0,unsigned int n=VALUE):Identifier(s, n|VALUE)
			{ v=val; };
		double operator = (double x) { return (v=x); };
		virtual int evalue(double *r);
		virtual void compile();
};

//*** Opzioni visualizzazione

enum VIDMODE  { INFISSA, RPN };  //metodo di scrittura dell'espressione
extern VIDMODE vidmode;

//*** Classe di interfaccia Expression

struct node {
	Identifier *link;
	node *next;
	~node() { delete next; };
};

class Expression: public Base
{
	char *s;
	char *sold;
	Object *ptr;
	func funcpoint;
	node* firstidentifier;
	virtual ostream& out(ostream &out) { return out << *ptr; };
	Identifier *Expression::ident_expr(int prior, int type);
	Object *base_expr();
	Object *MakeTree (int prior);

  public:
	Expression (char *s, Value *var=NULL);
	virtual ~Expression () { mydelete(ptr); delete firstidentifier;
							 delete funcpoint; };
	void maketree ();
	int installide (Identifier *p);
	virtual int installdefault();
	int simplify_length() { return ptr->simplify_length(); };
	void simplify_const();
	void onlycompile();
	void compile();
	double operator ()();
};

#endif